// @file vl_nnpool.cu
// @brief Pooling block MEX wrapper
// @author Andrea Vedaldi
// @author Karel Lenc

/*
Copyright (C) 2014-15 Andrea Vedaldi and Karel Lenc.
All rights reserved.

This file is part of the VLFeat library and is made available under
the terms of the BSD license (see the COPYING file).
*/

#include "bits/mexutils.h"
#include "bits/datamex.hpp"
#include "bits/nnpooling_hae.hpp"

#if ENABLE_GPU
#include "bits/datacu.hpp"
#endif

#include <assert.h>

/* option codes */
enum {
  opt_stride = 0,
  opt_pad,
  opt_method,
  opt_depth,
  opt_verbose,
  opt_cudnn,
  opt_no_cudnn,
} ;

/* options */
vlmxOption  options [] = {
  {"Stride",           1,   opt_stride            },
  {"Pad",              1,   opt_pad               },
  {"Method",           1,   opt_method            },
  {"Depth",            1,   opt_depth            },
  {"Verbose",          0,   opt_verbose           },
  {"CUDNN",            0,   opt_cudnn             },
  {"NoCUDNN",          0,   opt_no_cudnn          },
  {0,                  0,   0                     }
} ;

/* ---------------------------------------------------------------- */
/*                                                          Context */
/* ---------------------------------------------------------------- */

vl::MexContext context ;

/*
 Resetting the context here resolves a crash when MATLAB quits and
 the ~Context function is implicitly called on unloading the MEX file.
 */
void atExit()
{
  context.clear() ;
}

/* ---------------------------------------------------------------- */
/*                                                       MEX driver */
/* ---------------------------------------------------------------- */

enum {
  IN_DATA = 0, IN_SIZE, IN_INDICATOR, IN_DEROUTPUT, IN_END
} ;

enum {
  OUT_RESULT = 0, OUT_INDICATOR, OUT_END
} ;

void mexFunction(int nout, mxArray *out[],
                 int nin, mxArray const *in[])
{
  int poolWidth ;
  int poolHeight ;
  int strideX = 1 ;
  int strideY = 1 ;
  int padLeft = 0 ;
  int padRight = 0 ;
  int padTop = 0 ;
  int padBottom = 0 ;
  int inhibitDepth = 1;
  vl::PoolingMethod method = vl::vlPoolingMax ;
  bool backMode = false ;
  bool compute_indicator = false;

  int verbosity = 0 ;
  int opt ;
  int next = IN_END ;
  mxArray const *optarg ;
  mxGPUArray *mx_indicator = NULL;

  /* -------------------------------------------------------------- */
  /*                                            Check the arguments */
  /* -------------------------------------------------------------- */

  mexAtExit(atExit) ;

  if (nin < 2) {
    mexErrMsgTxt("The arguments are less than two.") ;
  }

  if(nin >=4 && !vlmxIsString(in[2],-1) && !vlmxIsString(in[3],-1)) {
    backMode = true;
  }
  else  {
    backMode = false;

    if(vlmxIsString(in[3],-1))
      next = 3;
    if(vlmxIsString(in[2], -1))
      next = 2;
  }
  if (nout == 2)
    compute_indicator = true;


  while ((opt = vlmxNextOption (in, nin, options, &next, &optarg)) >= 0) {
    switch (opt) {
      case opt_verbose :
        ++ verbosity ;
        break ;

      case opt_stride :
        if (!vlmxIsPlainMatrix(optarg,-1,-1)) {
          mexErrMsgTxt("STRIDE is not a plain matrix.") ;
        }
        switch (mxGetNumberOfElements(optarg)) {
          case 1:
            strideY = (int)mxGetPr(optarg)[0] ;
            strideX = strideY ;
            break ;
          case 2:
            strideY = (int)mxGetPr(optarg)[0] ;
            strideX = (int)mxGetPr(optarg)[1] ;
            break ;
          default:
            mexErrMsgTxt("STRIDE has neither one nor two elements.") ;
        }
        break ;

      case opt_pad :
        if (!vlmxIsPlainMatrix(optarg,-1,-1)) {
          mexErrMsgTxt("PAD is not a plain matrix.") ;
        }
        switch (mxGetNumberOfElements(optarg)) {
          case 1:
            padLeft = (int)mxGetPr(optarg)[0] ;
            padRight = padLeft ;
            padTop = padLeft ;
            padBottom = padLeft ;
            break ;
          case 4:
            padTop = (int)mxGetPr(optarg)[0] ;
            padBottom = (int)mxGetPr(optarg)[1] ;
            padLeft = (int)mxGetPr(optarg)[2] ;
            padRight = (int)mxGetPr(optarg)[3] ;
            break ;
          default:
            mexErrMsgTxt("PAD has neither one nor four elements.") ;
        }
        break;

        case opt_depth:
        if (!vlmxIsPlainMatrix(optarg,-1,-1)) {
          mexErrMsgTxt("depth is not a plain matrix.") ;
        }
        switch (mxGetNumberOfElements(optarg)) {
          case 1:
          inhibitDepth = (int)mxGetPr(optarg)[0] ;
          break ;
          default:
          mexErrMsgTxt("depth must have one element.") ;
        }

        break;

      case opt_method :
        if (!vlmxIsString(optarg,-1)) {
           vlmxError(vlmxErrInvalidArgument, "METHOD is not a string.") ;
        }
        if (vlmxIsEqualToStringI(optarg, "max")) {
          method = vl::vlPoolingMax ;
        } else if (vlmxIsEqualToStringI(optarg, "avg")) {
          method = vl::vlPoolingAverage ;
        } else if (vlmxIsEqualToStringI(optarg, "inhibit")) {
          method = vl::vlPoolingInhibit ;
        }
        else {
          vlmxError(vlmxErrInvalidArgument, "METHOD is not a supported method.") ;
        }
        break;

      case opt_no_cudnn :
#if ENABLE_CUDNN
        context.getCudaHelper().setCudnnEnabled(false) ;
#endif
        break ;

      case opt_cudnn :
#if ENABLE_CUDNN
        context.getCudaHelper().setCudnnEnabled(true) ;
#endif
        break ;

      default:
        break ;
    }
  }

  vl::MexTensor data(context) ;
  vl::MexTensor derOutput(context) ;

  data.init(in[IN_DATA]) ;

  if(data.getMemoryType()!=vl::GPU) {
    mexErrMsgTxt("The first input is not GPU arrays.") ;
  }

  if (backMode) { derOutput.init(in[IN_DEROUTPUT]) ; }

  if (backMode && ! vl::areCompatible(data, derOutput)) {
    mexErrMsgTxt("DATA and DEROUTPUT are not both CPU or GPU arrays.") ;
  }

  if (!vlmxIsPlainMatrix(in[IN_SIZE],-1,-1)) {
    mexErrMsgTxt("SIZE is not a plain matrix.") ;
  }
  switch (mxGetNumberOfElements(in[IN_SIZE])) {
    case 1:
      poolHeight = mxGetPr(in[IN_SIZE])[0] ;
      poolWidth = poolHeight ;
      break ;
    case 2:
      poolHeight = mxGetPr(in[IN_SIZE])[0] ;
      poolWidth = mxGetPr(in[IN_SIZE])[1] ;
      break ;
    default:
      mexErrMsgTxt("SIZE has neither one nor two elements.") ;
  }

  /* Basic compatibility of geometry */
  if (strideX < 1 || strideY < 1) {
    mexErrMsgTxt("At least one element of STRIDE is smaller than one.") ;
  }
  if (poolHeight == 0 || poolWidth == 0) {
    mexErrMsgTxt("A dimension of the pooling SIZE is void.") ;
  }
  if (data.getHeight() + (padTop+padBottom) < poolHeight ||
      data.getWidth() + (padLeft+padRight) < poolWidth) {
    mexErrMsgTxt("The pooling window is larger than the DATA (including padding).") ;
  }
  if (padLeft < 0 ||
      padRight < 0 ||
      padTop < 0 ||
      padBottom < 0) {
    mexErrMsgTxt("An element of PAD is negative.") ;
  }
  if (padLeft >= poolWidth ||
      padRight >= poolWidth ||
      padTop >= poolHeight  ||
      padBottom >= poolHeight) {
    mexErrMsgTxt("A padding value is larger or equal than the size of the pooling window.") ;
  }

   vl::TensorGeometry outputGeom;
  /* Get the output geometry */
  if(method == vl::vlPoolingInhibit) {
    outputGeom.setHeight(data.getHeight());
    outputGeom.setWidth(data.getWidth());
    outputGeom.setDepth(data.getDepth()) ;
    outputGeom.setSize(data.getSize()) ;
  }else {
    outputGeom.setHeight((data.getHeight() + (padTop+padBottom) - poolHeight)/strideY + 1);
    outputGeom.setWidth((data.getWidth()  + (padLeft+padRight) - poolWidth)/strideX + 1);
    outputGeom.setDepth(data.getDepth()) ;
    outputGeom.setSize(data.getSize()) ;
  }

  if (backMode && (derOutput != outputGeom)) {
    mexErrMsgTxt("DEROUTPUT dimensions are incompatible with X and POOL.") ;
  }

  /* Create output buffers */
  vl::Device type = data.getMemoryType() ;
  vl::MexTensor output(context) ;
  vl::MexTensor derData(context) ;
  vl::MexTensor derFilters(context) ;
  vl::MexTensor derBiases(context) ;

  if (!backMode) {
    output.init(type, outputGeom, 0) ;
  } else {
    derData.init(type, data.getGeometry(), 0) ;
  }

  // assign indicator
  if(compute_indicator) {
    mwSize dimensions [4] = {(mwSize)outputGeom.getHeight(),
     (mwSize)outputGeom.getWidth(),
     (mwSize)outputGeom.getDepth(),
     (mwSize)outputGeom.getSize()} ;

     if(method == vl::vlPoolingInhibit) {
      mx_indicator = mxGPUCreateGPUArray(4,
        dimensions,
        mxLOGICAL_CLASS,
        mxREAL,
        MX_GPU_DO_NOT_INITIALIZE);
        padLeft = inhibitDepth;
     } else {
      mx_indicator = mxGPUCreateGPUArray(4,
        dimensions,
        mxINT32_CLASS,
        mxREAL,
        MX_GPU_DO_NOT_INITIALIZE);
    }
  }
  else {
    mx_indicator = mxGPUCopyFromMxArray(in[IN_INDICATOR]);
  }


  if (verbosity > 0) {
    mexPrintf("vl_nnpool: %s; %s", backMode?"backward":"forward", (data.getMemoryType()==vl::GPU) ? "GPU" : "CPU") ;
    if (data.getMemoryType() == vl::GPU) {
#if ENABLE_CUDNN
      mexPrintf("; %s\n", context.getCudaHelper().getCudnnEnabled() ? "cuDNN" : "MatConvNet") ;
#else
      mexPrintf("; MatConvNet\n") ;
#endif
    } else {
      mexPrintf("; MatConvNet\n") ;
    }
    mexPrintf("vl_nnpool: stride: [%d %d], pad: [%d %d %d %d]\n",
              strideY, strideX,
              padTop, padBottom, padLeft, padRight) ;
    vl::print("vl_nnpool: data: ", data) ;
    mexPrintf("vl_nnpool: pooling: %d x %d\n", poolHeight, poolWidth);
    switch(method)
    {
      case vl::vlPoolingMax:
        mexPrintf("vl_nnpool: method: max\n") ;
      break;

      case vl::vlPoolingAverage:
        mexPrintf("vl_nnpool: method: average\n") ;
      break;

      case vl::vlPoolingInhibit:
        mexPrintf("vl_nnpool: method: inhibit\n") ;
      break;

    }
    if (backMode) {
      vl::print("vl_nnpool: derOutput: ", derOutput) ;
      vl::print("vl_nnpool: derData: ", derData) ;
    } else {
      vl::print("vl_nnpool: output: ", output) ;
    }
  }

  /* -------------------------------------------------------------- */
  /*                                                    Do the work */
  /* -------------------------------------------------------------- */
  vl::Error error ;
  if(method == vl::vlPoolingInhibit) {
    if (!backMode) {
      error = vl::nnpooling_forward_hae(context,
                                    output, data,
                                    method,
                                    poolHeight, poolWidth,
                                    strideY, strideX,
                                    padTop, padBottom, padLeft, padRight,
                                    (bool *)(mxGPUGetData(mx_indicator)), compute_indicator) ;
    } else {
      error = vl::nnpooling_backward_hae(context,
                                     derData, data, derOutput,
                                     method,
                                     poolHeight, poolWidth,
                                     strideY, strideX,
                                     padTop, padBottom, padLeft, padRight,
                                     (bool *)(mxGPUGetData(mx_indicator)), compute_indicator) ;
    }
  }else{
    if (!backMode) {
      error = vl::nnpooling_forward_hae(context,
                                    output, data,
                                    method,
                                    poolHeight, poolWidth,
                                    strideY, strideX,
                                    padTop, padBottom, padLeft, padRight,
                                    (int *)(mxGPUGetData(mx_indicator)), compute_indicator) ;
    } else {
      error = vl::nnpooling_backward_hae(context,
                                     derData, data, derOutput,
                                     method,
                                     poolHeight, poolWidth,
                                     strideY, strideX,
                                     padTop, padBottom, padLeft, padRight,
                                     (int *)(mxGPUGetData(mx_indicator)), compute_indicator) ;
    }
  }

  /* -------------------------------------------------------------- */
  /*                                                         Finish */
  /* -------------------------------------------------------------- */

  if (error != vl::vlSuccess) {
    mexErrMsgTxt(context.getLastErrorMessage().c_str()) ;
  }
  if (backMode) {
    out[OUT_RESULT] = derData.relinquish() ;
  } else {
    out[OUT_RESULT] = output.relinquish() ;
  }
  if (compute_indicator) {
    out[OUT_INDICATOR] = mxGPUCreateMxArrayOnGPU(mx_indicator);
  }
  mxGPUDestroyGPUArray(mx_indicator);
}
